<?php

// autoload_files.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'b4b4f2eb2151c57cadbd5714a7aba8b7' => $baseDir . '/foo.php',
    '99b24fc198db06c1d2d8118a8a5475eb' => $baseDir . '/bar.php',
);
