<?php

// autoload_namespaces.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = $vendorDir;

return array(
    'Main' => array($vendorDir . '/src'),
    'Lala' => array($vendorDir . '/src'),
);
