<?php

// autoload_namespaces.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'B\\Sub\\Name' => array($vendorDir . '/b/b/src'),
    'A\\B' => array($vendorDir . '/a/a/lib'),
    'A' => array($vendorDir . '/a/a/src'),
);
