<?php

// include_paths.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    $baseDir . '/lib',
    $baseDir . '/src',
    $vendorDir . '/a/a/lib1',
    $vendorDir . '/a/a/src1',
    $vendorDir . '/b/b/lib2',
    $vendorDir . '/c/c/lib3',
);
