<?php

// autoload_psr4.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'Acme\\Fruit\\' => array($baseDir . '/src-fruit'),
    'Acme\\Cake\\' => array($baseDir . '/src-cake', $baseDir . '/lib-cake'),
);
