<?php

// autoload_static.php @generated by Composer

namespace Composer\Autoload;

class ComposerStaticInit46489d6835a727203ffccd612295440e
{
    public static $files = array (
        '320cde22f66dd4f5d3fd621d3e88b98f' => __DIR__ . '/..' . '/symfony/polyfill-ctype/bootstrap.php',
    );

    public static $prefixLengthsPsr4 = array (
        'S' => 
        array (
            'Symfony\\Polyfill\\Ctype\\' => 23,
            'Symfony\\Component\\Filesystem\\' => 29,
        ),
    );

    public static $prefixDirsPsr4 = array (
        'Symfony\\Polyfill\\Ctype\\' => 
        array (
            0 => __DIR__ . '/..' . '/symfony/polyfill-ctype',
        ),
        'Symfony\\Component\\Filesystem\\' => 
        array (
            0 => __DIR__ . '/..' . '/symfony/filesystem',
        ),
    );

    public static $classMap = array (
        'Composer\\InstalledVersions' => __DIR__ . '/..' . '/composer/InstalledVersions.php',
        'Hooks' => __DIR__ . '/../..' . '/Hooks.php',
        'PluginA' => __DIR__ . '/..' . '/plugin/a/PluginA.php',
        'PluginB' => __DIR__ . '/..' . '/plugin/b/PluginB.php',
    );

    public static function getInitializer(ClassLoader $loader)
    {
        return \Closure::bind(function () use ($loader) {
            $loader->prefixLengthsPsr4 = ComposerStaticInit46489d6835a727203ffccd612295440e::$prefixLengthsPsr4;
            $loader->prefixDirsPsr4 = ComposerStaticInit46489d6835a727203ffccd612295440e::$prefixDirsPsr4;
            $loader->classMap = ComposerStaticInit46489d6835a727203ffccd612295440e::$classMap;

        }, null, ClassLoader::class);
    }
}
