<?php

// autoload_files.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    '5e70d6595c54512c151823ca0663ab51' => $baseDir . '/root.php',
);
