<?php

// include_paths.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    $vendorDir . '/a/a/lib',
    $vendorDir . '/b/b/library',
    $vendorDir . '/c/library',
);
