<?php

// autoload_files.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    '25b429360a61ca2629fdf9a4f484981c' => $baseDir . '/foo.php',
    '524f65941cc9a0fa65ff0ec097ccde8a' => $baseDir . '/bar.php',
);
